import configparser
import pathlib


def get_mappings(model_name: str) -> dict:
    mappings_path = pathlib.Path(__file__).parent / "mappings.ini"
    config = configparser.ConfigParser()
    config.read(mappings_path)
    return dict(config.items(model_name))



from py_apple_books.db.clause import Where
from typing import List, Optional, Any, Union, Dict
from py_apple_books.db.client import DBClient


class QueryCompiler:
    """SQL query compiler"""

    def __init__(self, client: DBClient):
        self.client = client

    def execute(self, query: str) -> list[Any]:
        return self.client.execute(query)


class Query:
    """SQL query builder"""

    # TODO: add join query support (single and multi-level)

    @staticmethod
    def select(table_name: str,
               fields: Union[List[str], str] = '*',
               where: Optional[List[Where]] = None,
               order_by: Optional[str] = None,
               limit: Optional[int] = None,
               use_or: bool = False) -> str:
        """
        Build and execute a SELECT query

        Args:
            table_name: The name of the table
            fields: The fields to select
            where: The WHERE clause
            order_by: The ORDER BY clause
            limit: The LIMIT clause
        """
        if isinstance(fields, list):
            fields_str = ', '.join(fields)
        else:
            fields_str = fields

        query = f"SELECT {fields_str} FROM {table_name}"

        if where:
            where_clauses = [str(clause) for clause in where]
            if use_or:
                query += f" WHERE {' OR '.join(where_clauses)}"
            else:
                query += f" WHERE {' AND '.join(where_clauses)}"

        if order_by:
            query += f" ORDER BY {order_by}"

        if limit:
            query += f" LIMIT {limit}"

        return query

    @staticmethod
    def insert(table_name: str,
               data: Dict[str, Any]) -> str:
        """
        Build and execute an INSERT query
        """
        fields = ', '.join(data.keys())
        values = ', '.join([f"'{value}'" for value in data.values()])
        query = f"INSERT INTO {table_name} ({fields}) VALUES ({values})"
        return query

    @staticmethod
    def update(table_name: str,
               data: Dict[str, Any],
               where: List[Where]) -> str:
        """
        Build and execute an UPDATE query
        """
        set_clause = ', '.join([f"{field} = '{value}'" for field, value in data.items()])
        where_clause = ' AND '.join([str(clause) for clause in where])
        query = f"UPDATE {table_name} SET {set_clause} WHERE {where_clause}"
        return query

    @staticmethod
    def delete(table_name: str,
               where: List[Where]) -> str:
        """
        Build and execute a DELETE query
        """
        where_clause = ' AND '.join([str(clause) for clause in where])
        query = f"DELETE FROM {table_name} WHERE {where_clause}"
        return query



from dataclasses import dataclass
from py_apple_books.models.base import Model
from datetime import datetime
from enum import Enum


class AnnotationColor(Enum):
    GREEN = 1
    BLUE = 2
    YELLOW = 3
    PINK = 4
    PURPLE = 5


@dataclass
class Annotation(Model):
    """
    Represents an annotation in the Apple Books library.
    """
    # Identifiers
    id: str
    asset_id: str

    # Status
    is_deleted: bool

    # Dates
    creation_date: datetime
    modification_date: datetime

    # Annotation details
    representative_text: str
    selected_text: str
    note: str
    is_underline: bool
    style: int

    # Location
    chapter: str
    location: str

    # Color
    color: str = None

    def __post_init__(self):
        """
        Converts the creation_date and modification_date from timestamp to datetime.
        """
        self.creation_date = datetime.fromtimestamp(self.creation_date) if self.creation_date else None
        self.modification_date = datetime.fromtimestamp(self.modification_date) \
            if self.modification_date else None

        if self.style in AnnotationColor._value2member_map_:
            self.color = AnnotationColor(self.style).name

    def __str__(self):
        return f"ID: {self.id}\nRepresentative text: {self.representative_text}\nSelected text: {self.selected_text}\nNote: {self.note}"



from typing import Any
import pathlib
import configparser
from py_apple_books.models.manager import ModelManager
from py_apple_books.models.relations import OneToMany, OneToOne, ManyToMany


class ModelBase(type):
    def __new__(mcs, name, bases, attrs):
        cls = super().__new__(mcs, name, bases, attrs)
        setattr(cls, 'manager', ModelManager(cls))

        cls.relations = []
        for attribute, value in attrs.items():
            if isinstance(value, OneToMany) or isinstance(value, OneToOne):
                relation_type = value.__class__.__name__
                forward_relation = {
                    'name': attribute,
                    'type': relation_type,
                    'related_model': value.related_model,
                    'foreign_key': value.foreign_key,
                }
                cls.relations.append(forward_relation)

                related_model = value.related_model
                if not hasattr(related_model, 'relations'):
                    related_model.relations = []
                backward_relation = {
                    'name': value.related_name,
                    'type': 'ManyToOne' if relation_type == 'OneToMany' else relation_type,
                    'related_model': cls,
                    'foreign_key': value.foreign_key,
                }
                related_model.relations.append(backward_relation)

            elif isinstance(value, ManyToMany):
                relation_type = value.__class__.__name__
                forward_relation = {
                    'name': attribute,
                    'type': relation_type,
                    'related_model': value.related_model,
                    'from_key': value.from_key,
                    'to_key': value.to_key,
                    'join_table': value.join_table,
                }
                cls.relations.append(forward_relation)

                related_model = value.related_model
                if not hasattr(related_model, 'relations'):
                    related_model.relations = []
                backward_relation = {
                    'name': value.related_name,
                    'type': relation_type,
                    'related_model': cls,
                    'from_key': value.to_key,
                    'to_key': value.from_key,
                    'join_table': value.join_table,
                }
                related_model.relations.append(backward_relation)

        return cls


class Model(metaclass=ModelBase):
    """
    Base class for all Apple Books models.
    """

    @classmethod
    def _get_mappings(cls, section: str, keys: list[str] | None = None) -> dict:
        mappings_path = pathlib.Path(__file__).parent / "mappings.ini"
        config = configparser.ConfigParser()
        config.read(mappings_path)
        if keys is None:
            return dict(config.items(section))
        return {key: config.get(section, key) for key in keys}

    @classmethod
    def from_db(cls, db_data: list[Any]) -> 'Model':
        fields_len = len(cls._get_mappings(cls.__name__))
        obj = cls(**dict(zip(cls._get_mappings(cls.__name__).keys(), db_data[:fields_len])))
        cls.manager.handle_relations(obj)
        return obj

    @classmethod
    def to_db(cls) -> dict:
        pass

    # TODO
    def save(self) -> str:
        """
        Saves the model to the database.
        """
        pass



from py_apple_books.models.book import Book
from py_apple_books.models.collection import Collection
from py_apple_books.models.annotation import Annotation, AnnotationColor

__all__ = ["Book", "Collection", "Annotation", "AnnotationColor"]



from dataclasses import dataclass
from py_apple_books.models.base import Model
from py_apple_books.models.annotation import Annotation
from py_apple_books.models.relations import OneToMany
import pathlib
from datetime import datetime


@dataclass
class Book(Model):
    """
    Represents a book in the Apple Books library.
    """
    id: int
    asset_id: str

    # Basic book information
    title: str
    author: str
    description: str
    genre: str
    content_type: str
    page_count: int

    # File information
    path: pathlib.Path
    filesize: int

    # Reading progress
    is_finished: bool
    reading_progress: float
    duration: float

    # Dates
    creation_date: datetime
    finished_date: datetime
    last_opened_date: datetime
    purchased_date: datetime

    # Flags
    is_explicit: bool
    is_locked: bool
    is_ephemeral: bool
    is_hidden: bool
    is_sample: bool
    is_store_audiobook: bool

    # User interactions
    rating: int

    # Relations
    annotations = OneToMany(
        related_model=Annotation,
        related_name='book',
        foreign_key='asset_id'
    )

    def __post_init__(self):
        self.creation_date = datetime.fromtimestamp(float(self.creation_date) / 1000) if self.creation_date else None
        self.finished_date = datetime.fromtimestamp(float(self.finished_date) / 1000) if self.finished_date else None
        self.last_opened_date = datetime.fromtimestamp(float(self.last_opened_date) / 1000) \
            if self.last_opened_date else None
        self.purchased_date = datetime.fromtimestamp(float(self.purchased_date) / 1000) if self.purchased_date else None
        self.duration = float(self.duration) / 1000 if self.duration else None
        self.reading_progress = float(self.reading_progress) * 100 if self.reading_progress else None

    def __str__(self):
        return f"ID: {self.id}\nTitle: {self.title}\nAuthor: {self.author}\nDescription: {self.description}"



import sqlite3
from pathlib import Path
from py_apple_books.db.exceptions import DBError, DBConnectionError, DBQueryError


class DBClient:
    def _get_sqlite_file(self, path: Path) -> Path:
        return list(path.glob("*.sqlite"))[0]

    def _get_cursor(self, paths: list[tuple[str, Path]]):
        _, first_path = paths[0]
        try:
            conn = sqlite3.connect(self._get_sqlite_file(first_path))
            cursor = conn.cursor()
            for db_name, path in paths[1:]:
                cursor.execute(f"ATTACH DATABASE '{self._get_sqlite_file(path)}' AS {db_name}")
            return cursor
        except sqlite3.Error as e:
            raise DBConnectionError(f"Error connecting to database: {e}")
        except IndexError:
            raise DBConnectionError("No sqlite files found. Please open iBooks at least once.")
        except Exception as e:
            raise DBError(f"Unexpected error while connecting to database: {e}")

    def execute(self, *args, **kwargs):
        raise NotImplementedError

    def close(self):
        self.cursor.close()


class AppleBooksDBClient(DBClient):
    """
    Wrapper for the Apple Books SQLite database
    """
    book_lib_db = ("lib_db", Path.home() / "Library/Containers/com.apple.iBooksX/Data/Documents/BKLibrary")
    anno_db = ("anno_db", Path.home() / "Library/Containers/com.apple.iBooksX/Data/Documents/AEAnnotation")

    def __init__(self):
        self.cursor = self._get_cursor(paths=[self.book_lib_db, self.anno_db])

    def execute(self, query: str) -> list:
        try:
            self.cursor.execute(query)
            return self.cursor.fetchall()
        except sqlite3.Error as e:
            raise DBQueryError(f"Error executing query: {e}")
        except Exception as e:
            raise DBError(f"Unexpected error while executing query: {e}")



from py_apple_books.models import Book, Collection, Annotation, AnnotationColor
from py_apple_books.models.manager import ModelIterable


class PyAppleBooks:
    """Facade class for accessing Apple Books data."""

    # -- collection actions --
    def list_collections(self, limit: int = None, order_by: str = None) -> ModelIterable:
        """List all collections."""
        return Collection.manager.all(limit=limit, order_by=order_by)

    def get_collection_by_id(self, collection_id: str) -> Collection:
        """Get a collection and its books."""
        return Collection.manager.filter(id=collection_id)[0]

    def get_collection_by_title(self, title: str) -> ModelIterable:
        """Get a collection and its books."""
        return Collection.manager.filter(title__contains=title)

    # -- book actions --
    def list_books(self, limit: int = None, order_by: str = None) -> ModelIterable:
        """List all books."""
        return Book.manager.all(limit=limit, order_by=order_by)

    def get_book_by_id(self, book_id: str) -> Book:
        """Get a book and its annotations."""
        return Book.manager.filter(id=book_id)[0]

    def get_book_by_title(self, title: str) -> ModelIterable:
        """Get a book by title."""
        return Book.manager.filter(title__contains=title)

    # -- annotation actions --
    def list_annotations(self, limit: int = None, order_by: str = None) -> ModelIterable:
        """List all annotations."""
        return Annotation.manager.all(limit=limit, order_by=order_by)

    def get_annotation_by_id(self, annotation_id: str) -> Annotation:
        """Get an annotation by id."""
        return Annotation.manager.filter(id=annotation_id)[0]

    def get_annotations_by_color(self, color: str, limit: int = None, order_by: str = None) -> ModelIterable:
        """Get annotations by color."""
        style = AnnotationColor[color.upper()].value
        return Annotation.manager.filter(style=style, limit=limit, order_by=order_by)

    def search_annotation_by_highlighted_text(self, text: str,
                                              limit: int = None, order_by: str = None) -> ModelIterable:
        """Search for annotations by highlighted text."""
        return Annotation.manager.filter(selected_text__contains=text, limit=limit, order_by=order_by)

    def search_annotation_by_note(self, note: str, limit: int = None, order_by: str = None) -> ModelIterable:
        """Search for annotations by note."""
        return Annotation.manager.filter(note__contains=note, limit=limit, order_by=order_by)

    def search_annotation_by_text(self, text: str, limit: int = None, order_by: str = None) -> ModelIterable:
        """Search for annotations by any text that contains the given text."""
        return Annotation.manager.filter(selected_text__contains=text,
                                         representative_text__contains=text,
                                         note__contains=text,
                                         use_or=True,
                                         limit=limit,
                                         order_by=order_by)



from typing import TypeVar, Type

T = TypeVar('T')


# TODO: add validations
class Relation:
    def __init__(self, related_model: Type[T], related_name: str):
        self.related_model = related_model
        self.related_name = related_name


class OneToMany(Relation):
    def __init__(self, related_model: Type[T], related_name: str, foreign_key: str):
        super().__init__(related_model, related_name)
        self.foreign_key = foreign_key


class OneToOne(Relation):
    def __init__(self, related_model: Type[T], related_name: str, foreign_key: str):
        super().__init__(related_model, related_name)
        self.foreign_key = foreign_key


class ManyToMany(Relation):
    def __init__(self, related_model: Type[T], related_name: str, from_key: str, to_key: str, join_table: str):
        super().__init__(related_model, related_name)
        self.from_key = from_key
        self.to_key = to_key
        self.join_table = join_table



class Clause:
    def __str__(self):
        pass

    def _escape_like(self, value: str) -> str:
        escaped_value = value.replace('%', r'\%').replace('_', r'\_')
        return f"'{escaped_value}'"

    def _escape_value(self, value) -> str:
        if isinstance(value, str):
            return f"'{value}'"
        return str(value)


class Join(Clause):
    def __init__(self, table: str, on: str, type: str = ''):
        self.table = table
        self.on = on
        self.type = type

    def __str__(self):
        return f"{self.type} JOIN {self.table} ON {self.on}"


class Where(Clause):
    def __init__(self, field: str, value, operator: str = '='):
        self.field = field
        self.value = value
        self.operator = operator
        self.is_list = isinstance(value, (list, tuple))

    def __str__(self):
        if self.operator == 'IN':
            # Handle lists for IN operator
            if isinstance(self.value, str):
                # If it's already a comma-separated string, use it directly
                formatted_value = f"({self.value})"
            else:
                # Format each item separately and join them
                formatted_items = [self._escape_value(item) for item in self.value]
                formatted_value = f"({', '.join(formatted_items)})"
            return f"{self.field} {self.operator} {formatted_value}"
        else:
            # Handle regular operators
            return f"{self.field} {self.operator} {self._escape_value(self.value)}"



from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

setup(
    name='py_apple_books',
    version='1.3.2',
    description='Python library for Apple Books',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Vignesh Iyer',
    author_email='vgnshiyer@gmail.com',
    packages=find_packages(),
    license='MIT',
    url='https://github.com/vgnshiyer/py-apple-books',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
    ],
    package_data={
        'py_apple_books': ['*.ini', 'models/*.ini'],
    },
)



from dataclasses import dataclass
from py_apple_books.models.base import Model
from py_apple_books.models.book import Book
from py_apple_books.models.relations import ManyToMany


@dataclass
class Collection(Model):
    """
    Represents a collection in the Apple Books library.
    """
    # Identifiers
    id: str
    title: str

    # Status
    is_deleted: bool
    is_hidden: bool

    # Collection details
    details: str
    books = ManyToMany(
        related_model=Book,
        related_name='collections',
        from_key='id',
        to_key='asset_id',
        join_table='book_collection'
    )

    def __str__(self):
        return f"ID: {self.id}\nTitle: {self.title}\nDetails: {self.details}"



from py_apple_books.db import QueryCompiler
from py_apple_books.db import AppleBooksDBClient
from py_apple_books.db import Query
from py_apple_books.db.clause import Where
from typing import Callable, Any, List
from functools import partial


class ModelIterable:
    def __init__(self, callable: Callable[[], list[Any]], model_class):
        self.run_query = callable
        self.model_class = model_class

    def __iter__(self):
        results = self.run_query()
        for result in results:
            yield self.model_class.from_db(result)

    def __len__(self):
        results = self.run_query()
        return len(results)

    def __getitem__(self, index: int) -> Any:
        results = self.run_query()
        return self.model_class.from_db(results[index])


class ModelManager:
    def __init__(self, model_class):
        self.model_class = model_class
        self.compiler = QueryCompiler(AppleBooksDBClient())

        self.model_name = self.model_class.__name__
        self.table_name = None
        if self.model_name != 'Model':
            self.table_name = self.model_class._get_mappings('Tables')[self.model_name.lower()]

    def _create_callable(self, query: str) -> Callable[[], list[Any]]:
        return partial(self.compiler.execute, query)

    def _get_db_field(self, field: str) -> str:
        return self.model_class._get_mappings(self.model_name)[field]

    def _get_fields(self, only: List[str] = None) -> List[str]:
        fields = list(self.model_class._get_mappings(self.model_name).values())
        if only:
            fields = [field for field in fields if field in only]
        return fields

    def _format_order_by(self, order_by: str) -> str:
        if order_by:
            field = order_by[1:] if order_by.startswith('-') else order_by
            direction = ' DESC' if order_by.startswith('-') else ''
            order_by = self.model_class._get_mappings(self.model_name)[field] + direction
        return order_by

    def all(self, only: List[str] = None, limit: int = None, order_by: str = None) -> ModelIterable:
        fields = self._get_fields(only)
        order_by = self._format_order_by(order_by)
        query = Query.select(self.table_name, fields=fields, limit=limit, order_by=order_by)
        return ModelIterable(callable=self._create_callable(query), model_class=self.model_class)

    def filter(self, only: List[str] = None, use_or: bool = False, limit: int = None,
               order_by: str = None, **filters) -> ModelIterable:
        fields = self._get_fields(only)

        where_clauses = []
        for field, value in filters.items():
            if field.endswith('__contains'):
                where_clauses.append(Where(self._get_db_field(field.split('__')[0]), f'%{value}%', operator='LIKE'))
            elif field.endswith('__in'):
                where_clauses.append(Where(self._get_db_field(field.split('__')[0]), value, operator='IN'))
            elif field.endswith('__gt'):
                where_clauses.append(Where(self._get_db_field(field.split('__')[0]), value, operator='>'))
            elif field.endswith('__gte'):
                where_clauses.append(Where(self._get_db_field(field.split('__')[0]), value, operator='>='))
            elif field.endswith('__lt'):
                where_clauses.append(Where(self._get_db_field(field.split('__')[0]), value, operator='<'))
            else:
                where_clauses.append(Where(self._get_db_field(field), value, operator='='))

        order_by = self._format_order_by(order_by)
        query = Query.select(self.table_name, fields=fields, where=where_clauses, use_or=use_or,
                             limit=limit, order_by=order_by)
        return ModelIterable(callable=self._create_callable(query), model_class=self.model_class)

    def handle_relations(self, model_object):
        """
        Handle the relations for a model object.
        """
        for relation in self.model_class.relations:
            # handle one-to-many relations
            if relation['type'] == 'OneToMany':
                related_model = relation['related_model']
                foreign_key = relation['foreign_key']
                val = getattr(model_object, foreign_key)
                setattr(model_object, relation['name'],
                        related_model.manager.filter(**{foreign_key: val}))

            # handle many-to-one relations
            elif relation['type'] == 'ManyToOne':
                related_model = relation['related_model']
                foreign_key = relation['foreign_key']
                val = getattr(model_object, foreign_key)
                try:
                    setattr(model_object, relation['name'],
                            related_model.manager.filter(**{foreign_key: val})[0])
                except IndexError:
                    setattr(model_object, relation['name'], None)

            # handle one-to-one relations
            elif relation['type'] == 'OneToOne':
                related_model = relation['related_model']
                foreign_key = relation['foreign_key']
                val = getattr(model_object, foreign_key)
                try:
                    setattr(model_object, relation['name'],
                            related_model.manager.filter(**{foreign_key: val})[0])
                except IndexError:
                    setattr(model_object, relation['name'], None)

            # handle many-to-many relations
            elif relation['type'] == 'ManyToMany':
                related_ids = self.get_related_ids(model_object, relation)
                related_model = relation['related_model']
                setattr(model_object, relation['name'],
                        related_model.manager.filter(**{f"{relation['to_key']}__in": related_ids}))

    def get_related_ids(self, model_object, relation: dict) -> list[str]:
        """
        Get related IDs for many-to-many relationships.
        """
        # This method is a minor contrivance to avoid multi-level join
        join_table = self.model_class._get_mappings('Tables')[relation['join_table']]
        from_key = self.model_class._get_mappings(relation['join_table'])[relation['from_key']]
        to_key = self.model_class._get_mappings(relation['join_table'])[relation['to_key']]
        val = getattr(model_object, relation['from_key'])

        query = Query.select(join_table, fields=[to_key], where=[Where(from_key, val, operator='=')])
        related_ids = self.compiler.execute(query)
        return [row[0] for row in related_ids]

    # TODO
    def add(self, data: dict) -> str:
        pass

    def update(self, id: int, data: dict) -> str:
        pass

    def delete(self, id: int) -> str:
        pass
